
module Prelude where

  id : forall {a : Set} -> a -> a
  id x = x

  data Empty : Set where

  magic : forall {a : Set} -> Empty -> a
  magic ()

  record Unit : Set where

  unit : Unit
  unit = record {}

  data Pair (a b : Set) : Set where
    _,_ : a -> b -> Pair a b

  fst : forall {a b} -> Pair a b -> a
  fst (x , _) = x

  snd : forall {a b} -> Pair a b -> b
  snd (_ , y) = y

  postulate
    Level : Set
    zero  : Level
    suc   : Level → Level
    _⊔_   : Level → Level → Level

  {-#  COMPILED_TYPE Level ()      #-}
  {-#  COMPILED zero ()            #-}
  {-#  COMPILED suc  (\_ -> ())    #-}
  {-#  COMPILED _⊔_  (\_ _ -> ())  #-}

  {-#  BUILTIN LEVEL     Level  #-}
  {-#  BUILTIN LEVELZERO zero   #-}
  {-#  BUILTIN LEVELSUC  suc    #-}
  {-#  BUILTIN LEVELMAX  _⊔_    #-}

  data _==_ {i : Level} {a : Set i} (x : a) : a -> Set i where
    Refl : x == x

  {-# BUILTIN EQUALITY _==_ #-}
  {-# BUILTIN REFL Refl #-}


  infix 6 _==_

  sym : {a : Set} {x y : a} -> x == y -> y == x
  sym Refl = Refl

  cong : {i j : Level} {a : Set i} {b : Set j} {x y : a} -> (f : a -> b) -> x == y -> f x == f y
  cong f Refl = Refl

  trans : {a : Set} {x y z : a} -> x == y -> y == z -> x == z
  trans Refl Refl = Refl

  _<>_ : {a : Set} (x y : a) -> Set
  x <> y = (x == y) -> Empty

  data Exists (a : Set) (b : a -> Set) : Set where
    Witness : (x : a) -> b x -> Exists a b

  fsts : forall {a b} -> Exists a b -> a
  fsts (Witness x _) = x

  snds : forall {a b} -> (x : Exists a b) -> (b (fsts x))
  snds (Witness _ y) = y

  data Nat : Set where
    Zero : Nat
    Succ : Nat -> Nat

  _+_ : Nat -> Nat -> Nat
  Zero + m = m
  (Succ k) + n = Succ (k + n)

  succInj : forall {k k' : Nat} -> Succ k == Succ k' -> k == k'
  succInj Refl = Refl

  data Either (a b : Set) : Set where
    Inl : a -> Either a b
    Inr : b -> Either a b

  decEqNat : (x y : Nat) -> Either (x == y) (x <> y)
  decEqNat Zero Zero = Inl Refl
  decEqNat Zero (Succ _) = Inr \()
  decEqNat (Succ k) Zero = Inr \()
  decEqNat (Succ k) (Succ k') with decEqNat k k'
  decEqNat (Succ .k') (Succ k') | Inl Refl = Inl Refl
  decEqNat (Succ k) (Succ k') | Inr neq = Inr (\succKEq -> neq (succInj succKEq))


  data Maybe (a : Set) : Set where
    Nothing : Maybe a
    Just : a -> Maybe a

  justInj : {a : Set} {x y : a} -> Just x == Just y -> x == y
  justInj Refl = Refl

  justNothing : forall {a : Set} {x : a} -> Nothing == Just x -> Empty
  justNothing ()

  data Inspect {a : Set} (x : a) : Set where
    _by_ : (y : a) -> x == y -> Inspect x

  inspect : forall {a} (x : a) -> Inspect x
  inspect x = x by Refl

